# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Some functions for spectral analysis."""

import numpy as np
import scipy.fftpack as fftpack
import scipy.signal as sig


def spectrum(xdata, sample_rate=1, window='hamming', one_side=False,
             return_amplitude=True, extra_len=None, save_energy=False):
    """Return the Fourier spectrum of signal.

    Parameters
    ----------
    xdata: array_like
        Signal values
    sample_rate: float
        Sampling frequency (Hz)
    window: str
        Window.
    one_side: boolean
        If True, the one-side spectrum is calculated (default value is
        False)
    return_amplitude: boolean
        If True, the amplitude spectrum is calculated
    extra_len: int
        If the value is set, the signal is padded with zeros to the
        extra_len value.
    save_energy: boolean
        If True, the result of FFT has the same energy as signal.  If
        False, the X (spectrum) is multiplied to 2/len(xdata). Use False
        if you want to see the correct amplitude of components in
        spectrum.

    Returns
    -------
    : np.ndarray of complex numbers
        Spectrum
    : np.ndarray of floats
        Frequency values (Hz)
    """
    win = sig.get_window(window, len(xdata))
    x_faded = xdata * win * len(win)/sum(win)

    actual_len = len(xdata)
    if extra_len:
        actual_len = max(extra_len, actual_len)

    sp_comp = fftpack.fft(x_faded, actual_len)
    if not save_energy:
        sp_comp *= 2/len(xdata)
    freqs = np.fft.fftfreq(len(sp_comp), 1/sample_rate)

    if one_side:
        ind = freqs >= 0
        freqs = freqs[ind]
        sp_comp = sp_comp[ind]

    if return_amplitude:
        return abs(sp_comp), freqs

    return sp_comp, freqs


def stft(xdata, sample_rate=1, nseg=256,
         nstep=None, window='hamming', nfft=None, padded=False):
    """Return result of short-time fourier transform.

    Parameters
    ----------
    xdata: numpy.ndarray
        Signal.
    sample_rate: float
       Sampling frequency (Hz).
    nseg: int
        Length of segment (in samples).
    nstep: int
        Optional. Length of step (in samples). If not setted then
        equal to nseg//2.
    window: str
        Window.
    nfft: int
        Length of the FFT. If None or less than nseg, the FFT length
        is nseg.

    Returns
    -------
    : numpy.ndarray
        Result of STFT, two-side spectrums.
    """
    if not nstep:
        nstep = nseg//2
    x_copy = xdata.copy()
    if padded:
        actual_len = len(x_copy) + (nseg - len(x_copy) % nseg) % nseg
        zer = np.zeros(actual_len)
        zer[:len(x_copy)] = x_copy
        x_copy = zer

    specs = []
    for i in range(0, len(x_copy)-nseg + 1, nstep):
        seg = x_copy[i: i+nseg]
        spec = spectrum(seg, sample_rate,
                        extra_len=nfft, window=window, save_energy=True)[0]
        specs.append(spec)

    return np.array(specs)


def calc_specgram(xdata, sample_rate=1, tdata=None, nseg=256,
                  nstep=None, freq_bounds=None, extra_len=None):
    """Return spectrogram data prepared to further plotting.

    Parameters
    ----------
    xdata: array_like
        Signal values
    sample_rate: float
        Sampling frequency (Hz)
    tdata: array_like
        Time values (sec)
    nseg: integer
        Length of window (number of samples)
    nstep: integer
        Length of step between Fourier transforms
    freq_bounds: tuple of 2 float
        Bounds of showed band
    extra_len: integer
        Number of values using for fft

    Return
    ------
    : np.ndarray
        Array of spectrums
    : np.ndarray
        Time values
    """
    if len(xdata) < nseg:
        return [], []
    if tdata is None:
        tdata = np.linspace(0, (len(xdata)-1)*sample_rate, len(xdata))
    else:
        sample_rate = 1/(tdata[1] - tdata[0])
    if not nstep:
        nstep = nseg//2
    specs = 2*stft(xdata=xdata, sample_rate=sample_rate, nseg=nseg,
                   nstep=nstep, nfft=extra_len, padded=True)
    if freq_bounds:
        freqs = np.fft.fftfreq(len(specs[0]), 1/sample_rate)
        ind = (freqs >= freq_bounds[0]) & (freqs <= freq_bounds[1])
        specs = specs[:, ind]
    t_new = np.linspace(tdata[nseg-1], tdata[-1], len(specs))
    return np.transpose(specs), t_new



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the base class for online filters."""

from math import pi
from collections import deque
import numpy as np
from dsplab.activity import Activity

PI = pi
PI2 = 2 * PI


def unwrap_point(phi):
    """Unwrap angle (for signle value)."""
    if phi < -pi:
        return phi + (1 + int(phi / -pi)) * pi
    if phi > pi:
        return phi - (1 + int(phi / pi)) * pi
    return phi


class QueueFilter(Activity):
    """Online filter with queue.

    Parameters
    ----------
    ntaps: int
        Lenght of filter.
    fill_with: object
        Initial value of every element of queue.
    """
    def __init__(self, ntaps, fill_with=0):
        super().__init__()
        self.queue = deque([fill_with]*ntaps, maxlen=ntaps)
        self.ntaps = ntaps

    def __call__(self, sample):
        """Add sample to queue."""
        self.queue.append(sample)
        return self.proc_queue()

    def proc_queue(self):
        """Process queue."""
        raise NotImplementedError


class Delayer(QueueFilter):
    """Provide delay in online processing."""
    def proc_queue(self):
        return self.queue[0]


class And(Activity):
    """And operation."""
    def __call__(self, sample):
        """Do operation.

        Parameters
        ----------
        sample: array_like of floats
            Input values.
        """
        res = 1
        for value in sample:
            res *= value
        return res


class Or(Activity):
    """Universal online filter.

    Parameters
    ----------
    ntaps: int
        Length of internal queue using for accumulation of input
        samples. Default is None.
    smooth_ntaps: int
        Length of queue using for smoothing output values. Default
        id None.
    fill_with: object
        Initial value of every element of queues.
    step: int
        Step. Must be positive.
    """
    def __init__(self, ntaps=None, smooth_ntaps=None, fill_with=0, step=1):
        super().__init__()
        self.add_sample_func = None
        if (ntaps is None) and (smooth_ntaps is None):
            self.add_sample_func = self.__add_sample_simple
        elif (ntaps is not None) and (smooth_ntaps is None):
            self.add_sample_func = self.__add_sample_only_queue
        elif (ntaps is None) and (smooth_ntaps is not None):
            self.add_sample_func = self.__add_sample_only_smooth
        else:
            self.add_sample_func = self.__add_sample_full

        if ntaps is not None:
            self.queue = deque([fill_with]*ntaps, maxlen=ntaps)
        if smooth_ntaps is not None:
            self.smooth_queue = deque(
                [fill_with]*smooth_ntaps,
                maxlen=smooth_ntaps
            )
            wind = np.hamming(smooth_ntaps)
            self.wind = wind / sum(wind)

        self.step = step
        self.steps = 0

        self.ntaps = ntaps
        self.smooth_ntaps = smooth_ntaps

    def __call__(self, sample):
        """Add input sample to filter and return output value.

        Parameters
        ----------
        sample: object
            Input sample.

        Returns
        -------
        : object
            Output value.
        """
        return self.add_sample_func(sample)

    def __add_sample_simple(self, sample):
        """Add sample without using queues."""
        self.steps += 1
        if self.steps == self.step:
            self.steps = 0
            return self.proc_sample(sample)
        return None

    def __add_sample_only_queue(self, sample):
        """Add sample with no smoothing."""
        self.steps += 1
        self.queue.append(sample)
        if self.steps == self.step:
            self.steps = 0
            return self.proc_queue()
        return None

    def __add_sample_only_smooth(self, sample):
        """Add sample with not internal queue but with smoothed
        ouput."""
        self.steps += 1
        if self.steps == self.step:
            self.steps = 0
            self.smooth_queue.append(self.proc_sample(sample))
            resm = np.dot(np.array(self.smooth_queue), self.wind)
            return resm
        return None

    def __add_sample_full(self, sample):
        """Add sample with internal queue and smoothing of ouput
        values."""
        self.steps += 1
        self.queue.append(sample)
        if self.steps == self.step:
            self.steps = 0
            self.smooth_queue.append(self.proc_queue())
            resm = np.dot(np.array(self.smooth_queue), self.wind)
            return resm
        return None

    def proc_queue(self):
        """Process queue.

        Returns
        -------
        : object
            Ouput value.
        """

    def proc_sample(self, sample):
        """Process sample.

        Parameters
        ----------
        sample: object
            Input sample.

        Returns
        -------
        : object
            Output value.
        """



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements Prony decomposition of signal."""

import numpy as np
import numpy.linalg as linalg


def prony_decomp(xdata, ncomp):
    """Prony decomposition of signal.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    ncomp: integer
        Number of components. 2*ncomp must be less tham length of xdata.

    Returns
    -------
    : np.array
        Mu-values.
    : np.array
        C-values.
    : np.array
        Components.
    """
    samples_total = len(xdata)
    if 2*ncomp > samples_total:
        return None
    d_matrix = []
    for i in range(ncomp, samples_total):
        row = [xdata[i-j-1] for j in range(0, ncomp)]
        d_matrix.append(np.array(row))
    d_matrix = np.array(d_matrix)
    d_column = np.array([xdata[i] for i in range(ncomp, samples_total)])

    a = linalg.lstsq(d_matrix, d_column, rcond=None)[0]
    p = np.array([1] + [-ai for ai in a])
    mu_vals = np.roots(p)

    d_matrix = []
    for i in range(samples_total):
        row = [mu_vals[j]**i for j in range(ncomp)]
        d_matrix.append(np.array(row))
    d_matrix = np.array(d_matrix)
    d_column = np.array([xdata[i] for i in range(samples_total)])

    c_vals = linalg.lstsq(d_matrix, d_column, rcond=None)[0]

    components = []
    for i in range(0, ncomp):
        comp = [c_vals[i] * (mu_vals[i]**k) for k in range(samples_total)]
        components.append(np.array(comp).real)
    components = np.array(components)

    return mu_vals, c_vals, components



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Filtration of signals."""

import numpy as np
import scipy.signal as sig
from scipy.fftpack import fft, ifft


def _stupid_filter(xdata, fr_resp):
    """Filter signal using setted frequency response.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    fr_resp: np.array
        Frequency response of ideal filter.

    Returns
    -------
    : np.array
        Filteres signal.
    """
    spectrum = fft(xdata * sig.tukey(len(xdata)))
    res_xs = np.real(ifft(spectrum * fr_resp))
    return res_xs


def stupid_lowpass_filter(xdata, sample_rate, cutoff):
    """Return low-pass filtered signal.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    sample_rate: float
        Sampling frequency.
    cutoff: float
        Cutoff frequency.

    Returns
    -------
    : np.array
        Filteres signal.
    """
    num = len(xdata)
    fr_resp = np.zeros(num)
    freqs = np.fft.fftfreq(num, 1/sample_rate)
    fr_resp[abs(freqs) <= cutoff] = 1
    res_xs = _stupid_filter(xdata, fr_resp)
    return res_xs


def stupid_bandpass_filter(xdata, sample_rate, bandpass):
    """Return low-pass filtered signal.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    sample_rate: float
        Sampling frequency.
    bandpass: np.array of 2 floats
        Bounds of bandpass (Hz).

    Returns
    -------
    : np.array
        Filteres signal.
    """
    num = len(xdata)
    fr_resp = np.zeros(num)
    freqs = np.fft.fftfreq(num, 1 / sample_rate)
    fr_resp[(abs(freqs) >= bandpass[0]) & (abs(freqs) <= bandpass[1])] = 1
    res_xs = _stupid_filter(xdata, fr_resp)
    return res_xs


def butter_filter(xdata, sample_rate, freqs, order, btype='band'):
    """Butterworth filter.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    sample_rate: float
        Sampling frequency (Hz).
    freqs: array_like
        One or two frequencies.
    order: integer
        Order of filter.
    btype: str ('band' | 'lowpass')
        Type of filter.

    Returns
    -------
    : np.array
        filtered signal.
    """
    nyq = 0.5 * sample_rate
    freqs = np.array(freqs)
    freqs /= nyq
    b_coeffs, a_coeffs = sig.butter(order, freqs, btype=btype)
    res_xs = sig.lfilter(b_coeffs, a_coeffs, xdata)
    return res_xs


def find_butt_bandpass_order(band, sample_rate):
    """Calculate the order of Butterworth bandpass filter using
    minimization of metric between ideal and real frequency response.

    Parameters
    ----------
    band: array_like
        Pair of frequencies. Bounds of bandpass (Hz).
    sample_rate: float
        Sample rate (Hz).

    Returns
    -------
    : integer
        Order of filter.

    """
    spectrum_len = round(60 * 120 * sample_rate)
    unit_pulse = np.zeros(spectrum_len)
    unit_pulse[1] = 1
    ideal_fr = np.zeros(spectrum_len)
    freqs = np.fft.fftfreq(spectrum_len, 1/sample_rate)
    ideal_fr[(freqs >= band[0]) & (freqs <= band[1])] = 1
    ideal_fr = ideal_fr[:spectrum_len//2]
    prev_metric = np.inf
    for order in range(3, 21):
        impulse_response = butter_filter(
            unit_pulse,
            sample_rate,
            band,
            order,
            btype='band'
        )
        real_fr = abs(fft(impulse_response))[:spectrum_len//2]
        metric = np.sum((real_fr - ideal_fr)**2)**0.5
        best_order = order
        if (np.isnan(metric)) or (metric >= prev_metric):
            best_order -= 1
            break
        prev_metric = metric
    return best_order-1


def haar_one_step(xdata, tdata, denominator=2):
    """One cascade of Haar transform.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    tdata: array_like
        Time values.
    denominator: integer
        Denominator used in Haar transform (default is 2).

    Returns
    -------
    : np.array
        Scaled signal values.
    : np.array.
        Details of x
    : np.array.
        Decimated time values
    """
    scl = []
    det = []
    res_ts = []
    for x_left, x_right in zip(xdata[::2], xdata[1::2]):
        scl.append((x_left+x_right)/denominator)
        det.append((x_left-x_right)/denominator)
    res_ts = tdata[1::2]
    return np.array(scl), np.array(det), np.array(res_ts)


def haar_scaling(xdata, tdata, steps_number):
    """Scaling with Haar transform.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    tdata: array_like
        Time values.
    steps_number: integer
        Number of cascades.

    Returns
    -------
    : np.array
        Scaled signal values.
    : np.array
        Decimated time values.
    """
    res_xs = xdata.copy()
    res_ts = tdata.copy()
    i = 0
    while i < steps_number:
        res = haar_one_step(res_xs, res_ts, denominator=2)
        res_xs, res_ts = res[0], res[2]
        i += 1
    return res_xs, res_ts


def smooth(xdata, ntaps=3, cut=True):
    """Smooth signal with Hamming window."""
    wind = np.hamming(ntaps)
    wind = wind / sum(wind)
    res = sig.lfilter(wind, [1], xdata)
    if cut:
        res = res[ntaps:]
    return res


def trend_smooth(xdata, sample_rate=1, tdata=None, cut_off=0.5):
    """Calculate trend of signal using smoothing filter.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    tdata: array_like
        Time values.
    cut_off: float
        The frequencies lower than this are trend's frequencies.

    Returns
    -------
    : np.array
        Trend values.
    : np.array
        Time values.
    """
    x_len = len(xdata)

    if tdata is None:
        tdata = np.linspace(0, (x_len-1)*sample_rate, x_len)
    else:
        sample_rate = 1.0 / (tdata[1] - tdata[0])

    win_len = int(sample_rate / 2 / cut_off)
    if win_len >= x_len:
        return None
    trend_xs = smooth(xdata, win_len)
    trend_ts = tdata[win_len:].copy()
    return trend_xs, trend_ts



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helpers."""

import importlib
import json


def is_iterable(obj):
    """Check if obj is iterable."""
    try:
        [e for e in obj]
    except TypeError:
        return False
    return True


def import_entity(name):
    """Import class by name."""
    parts = name.split('.')
    module_name = '.'.join(parts[:-1])
    entity_name = parts[-1]
    if module_name == "":
        module_name = '__main__'
    module = importlib.import_module(module_name)
    entity = getattr(module, entity_name)
    return entity


def pretty_json(some_object):
    """Return json representation of object."""
    return json.dumps(some_object, sort_keys=True, indent=4,
                      separators=(',', ': '))



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Playing signal from file. Online mode for offline data."""

import threading
from threading import Lock, Event
from collections import deque
import time
import random
import logging

LOG = logging.getLogger(__name__)


class RepeatedTimer:
    """Timer."""
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self._interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = None
        self.lock = Lock()

    def set_interval(self, interval):
        """Set interval."""
        self._interval = interval

    def get_interval(self):
        """Get interval."""
        return self._interval

    interval = property(get_interval, set_interval,
                        doc="Timeout interval (sec)")

    def start(self):
        """Start timer."""
        self.next_call = time.time()
        self.is_running = True
        self._repeat()

    def stop(self):
        """Stop timer."""
        self._timer.cancel()
        self.is_running = False

    def _repeat(self):
        # LOG.debug("%s._repeat" % self.__class__.__name__)
        if not self.is_running:
            return
        with self.lock:
            self.next_call += self._interval
            self._timer = threading.Timer(
                self.next_call - time.time(),
                self._repeat
            )
            self._timer.start()
        self.function(*self.args, **self.kwargs)


class SignalPlayer:
    """Class for playing text file as stream."""
    def __init__(self, interval):
        self.interval = interval
        self.queue = deque([], maxlen=100)
        self.timer = RepeatedTimer(interval, self._produce_data)
        self.new_data_ready = Event()
        self.lock = Lock()
        self.data_producer = None

    def set_data_producer(self, data_producer):
        """Set adapter with get_sample() method."""
        self.data_producer = data_producer

    def start(self):
        """Start player."""
        # LOG.debug('call SignalPlayer.start()')
        self.new_data_ready.clear()
        self.timer.set_interval(self.interval)
        self.data_producer.start()
        self.timer.start()

    def stop(self):
        """Stop player."""
        # LOG.debug("%s.stop()" % self.__class__.__name__)
        self.timer.stop()
        self.data_producer.stop()

    def _produce_data(self):
        # LOG.debug("%s._produce_data()" % self.__class__.__name__)
        with self.lock:
            sample = self.data_producer.get_sample()
        self.queue.append(sample)
        self.new_data_ready.set()
        return sample

    def get_sample(self):
        """Return sample."""
        try:
            sample = self.queue.popleft()
        except IndexError:
            self.new_data_ready.clear()
            self.new_data_ready.wait()
            sample = self.queue.popleft()
        return sample


class DataProducer:
    """Base class for adapters for data producer."""
    def get_sample(self):
        """Return sample."""
        raise NotImplementedError

    def start(self):
        """Do some operations in producer when player starts."""

    def stop(self):
        """Do some operations in producer when player stops."""


class RandomDataProducer(DataProducer):
    """Data producer with random values on output."""
    def __init__(self, interval):
        self.interval = interval

    def get_sample(self):
        """Return sample."""
        sample = random.randint(*self.interval)
        return sample


class CsvDataProducer(DataProducer):
    """Produces sample from headered CSV file."""
    def __init__(self, file_name=None, delimiter=';',
                 encoding='utf-8', columns=None):
        self.file_name = file_name
        self.encoding = encoding
        self._delimiter = None
        self.set_delimiter(delimiter)
        self._keys = None
        if columns is not None:
            self.select_columns(columns)
        self._headers = None
        self._indexes = None
        self._lines = None

    def set_delimiter(self, delimiter):
        """Set delimiter."""
        self._delimiter = delimiter

    def get_delimiter(self):
        """Return delimiter."""
        return self._delimiter

    delimiter = property(get_delimiter, set_delimiter,
                         doc="delimiter in CSV file.")

    def set_file(self, file_name, delimiter=None, encoding='utf-8'):
        """Set file for reading."""
        self.file_name = file_name
        if delimiter is not None:
            self.set_delimiter(delimiter)
        self.encoding = encoding

    def select_columns(self, keys):
        """Select returned columns. Numbers or names of columns can be
        used."""
        self._keys = keys

    def _detect_indexes(self):
        """Detect indexes of selected columns."""
        if isinstance(self._keys[0], int):
            self._indexes = self._keys
        else:
            indexes = []
            for key in self._keys:
                try:
                    index = self._headers.index(key)
                    indexes.append(index)
                except ValueError:
                    pass
            self._indexes = indexes

    def start(self):
        """Init reader."""
        # LOG.debug('Call CsvDataProducer.start()')
        with open(self.file_name, encoding=self.encoding) as buf:
            self._lines = iter(buf.read().split('\n'))
        line = next(self._lines)
        self._headers = line.split(self._delimiter)
        self._detect_indexes()

    def get_sample(self):
        """Return sample."""
        sample = []
        try:
            line = next(self._lines)
            full_sample = line.split(self._delimiter)
            if self._indexes is None:
                return full_sample
            for ind in self._indexes:
                sample.append(full_sample[ind])
        except StopIteration:
            sample = ['' for ind in self._indexes]
        except IndexError:
            sample = ['' for ind in self._indexes]
        return sample



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the Node and Plan classes. Node can be
understood as the workplace for worker. Node can have inputs that are
also nodes. Plan is the system of linked nodes."""

import logging
from warnings import warn

from dsplab.activity import get_work_from_dict
from dsplab.activity import Activity

LOG = logging.getLogger(__name__)


class Node(Activity):
    """Base class for nodes."""
    def __init__(self, inputs=None):
        super().__init__()
        self._id = None
        self._inputs = []
        if inputs is not None:
            self._inputs = inputs
        self._res = None
        self._start_hook = None
        self._start_hook_args = None
        self._start_hook_kwargs = None
        self._stop_hook = None
        self._stop_hook_args = None
        self._stop_hook_kwargs = None

        self._res_info = None

    def set_id(self, value):
        """Set ID for node."""
        self._id = value

    def get_id(self):
        """Return ID of node."""
        return self._id

    node_id = property(get_id, set_id, doc="ID of node.")

    def get_inputs(self):
        """Return inputs."""
        return self._inputs

    def set_inputs(self, inputs):
        """Set inputs."""
        self._inputs = inputs

    inputs = property(get_inputs, set_inputs)

    def set_start_hook(self, func, *args, **kwargs):
        """Set start hook."""
        self._start_hook = func
        self._start_hook_args = args
        self._start_hook_kwargs = kwargs

    def set_stop_hook(self, func, *args, **kwargs):
        """Set stop hook."""
        self._stop_hook = func
        self._stop_hook_args = args
        self._stop_hook_kwargs = kwargs

    def run_start_hook(self):
        """Run function associated with start hook."""
        if self._start_hook is not None:
            self._start_hook(*self._start_hook_args,
                             **self._start_hook_kwargs)

    def run_stop_hook(self):
        """Run function associated with stop hook."""
        if self._stop_hook is not None:
            self._stop_hook(*self._stop_hook_args,
                            **self._stop_hook_kwargs)

    def is_output_ready(self):
        """Check if the calculation in the node is finished."""
        ans = self._res is not None
        return ans

    def clear_result(self):
        """Clear the result."""
        self._res = None

    def reset(self):
        """Deprecated."""
        warn("Node.reset() is deprecated. Use clear_result() instead.")
        self.clear_result()

    def is_inputs_ready(self):
        """Check if data in all inputs is ready."""
        for inpt in self._inputs:
            if not inpt.is_output_ready():
                return False
        return True

    def get_result(self):
        """Return the calculated data."""
        return self._res

    def set_result_info(self, info):
        """Appent to info the desctription of the output data."""
        self._res_info = info

    def get_result_info(self):
        """Return result info."""
        return self._res_info

    result_info = property(get_result_info, set_result_info,
                           doc="Information about result")

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class WorkNode(Node):
    """Node with work."""
    def __init__(self, work=None, inputs=None):
        super().__init__(inputs)
        self._work = None
        self._func = work
        self.set_work(work)

    def get_work(self):
        """Return work of the node."""
        return self._work

    def set_work(self, work):
        """Set work for the node."""
        self._work = work
        self._func = work

    work = property(get_work, set_work, doc="Work in node")

    def reduce_call(self):
        """Try to reduce call chain."""
        try:
            self._func = self._work.worker.__call__
        except AttributeError:
            pass

    def __call__(self, data):
        self._res = self._work(*data)


class MapNode(WorkNode):
    """Apply work to all components of iterable input and build
    iterable output."""
    def __call__(self, data):
        self._res = []

        if len(self._inputs) > 1:
            self._res = []
            for zipped_args in map(list, zip(*data)):
                res_part = self._work(*zipped_args)
                self._res.append(res_part)
        elif len(self._inputs) == 1:
            self._res = []
            for comp in data[0]:
                comp_res = self._work(comp)
                self._res.append(comp_res)
        else:
            raise RuntimeError('MapNode must have input.')


class SelectNode(Node):
    """Select component of output."""
    def __init__(self, index, inputs=None):
        super().__init__(inputs)
        self.index = index

    def __call__(self, data):
        if len(data) > 1:
            data_tr = list(map(list, zip(*data)))
            self._res = data_tr[self.index]
        elif len(data) == 1:
            self._res = data[0][self.index]
        else:
            raise RuntimeError('SelectNode must have input.')


class PackNode(Node):
    """Pack input to output."""
    def __call__(self, data=None):
        self._res = data


class PassNode(Node):
    """Pass input to ouput."""
    def __call__(self, data):
        self._res = data[0]


class Plan(Activity):
    """The plan. Plan is the system of linked nodes."""
    def __init__(self, descr=None, quick=False):
        super().__init__()
        self._nodes = []
        self._inputs = []
        self._outputs = []
        self._progress_func = None

        self._quick = None
        self.set_quick(quick)

        self._sequence = []

    def set_descr(self, text):
        """Set description of plan."""
        self._descr = text

    def get_descr(self):
        """Return description of plan."""
        return self._descr

    descr = property(get_descr, set_descr, doc="Description of plan")

    def set_quick(self, value=True):
        """Make plan quick (for online with no hooks) or not."""
        self._quick = value
        if not value:
            self._run_func = self.run
        else:
            self._run_func = self.quick_run

    def _detect_sequence(self):
        """Find sequence of nodes for execution."""
        self._sequence = []
        while True:
            finished = True
            for node in self._nodes:
                if (node in self._sequence) or (node in self._inputs):
                    continue
                if set(node.inputs) <= set(self._sequence) | set(self._inputs):
                    self._sequence.append(node)
                    finished = False
            if finished:
                break

    def add_node(self, node, inputs=None):
        """Add node to plan."""
        self._nodes.append(node)
        if inputs is not None:
            node.inputs = inputs
        self._detect_sequence()

    def remove_node(self, node):
        """Remove node from plan."""
        if node not in self._nodes:
            raise RuntimeError("No such node")
        for n in self._nodes:
            if node in n.inputs:
                n.inputs.remove(node)
        self._nodes.remove(node)
        self._detect_sequence()

    def clear(self):
        """Clear plan."""
        self._nodes = []
        self._inputs = []
        self._outputs = []
        self._sequence = []

    def get_outputs(self):
        """Return output nodes."""
        return self._outputs

    def set_outputs(self, outputs):
        """Set output nodes."""
        self._outputs = outputs

    outputs = property(get_outputs,
                       set_outputs,
                       doc="The nodes wich are outputs.")

    def get_inputs(self):
        """Return input nodes."""
        return self._inputs

    def set_inputs(self, inputs):
        """Set input nodes."""
        self._inputs = inputs
        self._detect_sequence()

    inputs = property(get_inputs,
                      set_inputs,
                      doc="The nodes which are inputs.")

    def get_nodes(self):
        """Return the list of nodes."""
        return self._nodes

    def set_progress_hook(self, func):
        """Set progress handler."""
        self._progress_func = func

    def run(self, data):
        """Run plan."""
        for node in self._nodes:
            node.clear_result()

        for [node, node_data] in zip(self._inputs, data):
            node.run_start_hook()
            node([node_data])
            node.run_stop_hook()
            if self._progress_func is not None:
                self._progress_func()

        while True:
            finished = True
            for node in self._nodes:
                if not node.is_output_ready() and node.is_inputs_ready():
                    finished = False
                    input_nodes = node.get_inputs()
                    node_data = []
                    for input_node in input_nodes:
                        node_data.append(input_node.get_result())
                    node.run_start_hook()
                    node(node_data)
                    node.run_stop_hook()
                    if self._progress_func is not None:
                        self._progress_func()
            if finished:
                break

        return [output.get_result() for output in self._outputs]

    def reduce_calls(self):
        """Reduce call chains for all nodes. Recommended before run
        quick plans."""
        for node in self._nodes:
            if isinstance(node, WorkNode):
                node.reduce_call()
            if isinstance(node, MapNode):
                node.reduce_call()

    def quick_run(self, data):
        """Sequential execution of plan with no hooks (for on-line
        quick processing)."""
        for node, node_data in zip(self._inputs, data):
            node([node_data])
        for node in self._sequence:
            node_data = []
            for input_node in node.inputs:
                node_data.append(input_node.get_result())
            node(node_data)
        return [output.get_result() for output in self._outputs]

    def verify(self):
        """Validate plan.

        Returns
        -------
        : bool
            True if success, False otherwise.
        : str
            Empty string or description of error.
        """
        if not self._inputs:
            return False, "There are no inputs in the plan."
        if not self._outputs:
            return False, "There are no outputs in the plan."
        return True, ""

    def __call__(self, data):
        """Run plan."""
        return self._run_func(data)


def get_plan_from_dict(settings, params=None):
    """Create and return instance of Plan described in dictionary.

    Parameters
    ----------
    setting: dict
        Dictionary with plan.
    params: dict
        Dictionary with parameters like "$name" for plan.

    Returns
    -------
    : Plan
        The instance of Plan.

    **Keys in settings**

    - 'descr' - description of the plan (optional)
    - 'nodes' - list of dicts with nodes settings
    - 'inputs' - list of inputs nodes ids
    - 'outputs' - list of output nodes ids

    **Common settings for nodes**

    - 'id' - id of node
    - 'inputs' - list of ids of input nodes for this node
    - 'result' - result description

    **Settings for WorkNode and MapNode**

    - 'work' - dict with work settings

    **Settings for PackNode**

    - 'index' - index of selected item
    """
    # LOG.debug("get_plan_from_dict()")
    plan = Plan()
    if 'descr' in settings:
        plan.set_descr(settings['descr'])

    nodes = {}
    nodes_settings = settings['nodes']
    for node_settings in nodes_settings:
        # LOG.debug('Node settings:{} '.format(node_settings))
        node_id = node_settings['id']
        # LOG.debug("Parse node with id=%s", node_id)

        try:
            node_class = node_settings['class']
        except KeyError:
            node_class = 'WorkNode'

        if node_class == 'WorkNode':
            node = WorkNode()
            work_settings = node_settings['work']
            work = get_work_from_dict(work_settings, params)
            node.work = work
        elif node_class == 'MapNode':
            node = MapNode()
            work_settings = node_settings['work']
            work = get_work_from_dict(work_settings, params)
            node.work = work
        elif node_class == 'PackNode':
            node = PackNode()
        elif node_class == 'SelectNode':
            index = node_settings['index']
            node = SelectNode(index)
        elif node_class == 'PassNode':
            node = PassNode()
        else:
            message = "Unsupported node class: {}".format(node_class)
            raise ValueError(message)

        if 'result' in node_settings:
            node.set_result_info(node_settings['result'])

        node.set_id(node_id)
        nodes[node_id] = node

    for node_settings in nodes_settings:
        node_id = node_settings['id']
        if 'inputs' in node_settings.keys():
            inputs = [nodes[key] for key in node_settings['inputs']]
            plan.add_node(nodes[node_id], inputs=inputs)
        else:
            plan.add_node(nodes[node_id])

    if 'inputs' in settings:
        inputs = [nodes[key] for key in settings['inputs']]
        plan.set_inputs(inputs)

    if 'outputs' in settings:
        outputs = [nodes[key] for key in settings['outputs']]
        plan.set_outputs(outputs)

    return plan



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Functions for modulation and demodulation."""

from math import pi, cos
import numpy as np
from numpy import unwrap, angle, diff
import scipy.signal as sig


PI = pi
_2PI = 2 * PI


def harm(length, sample_rate, amp, freq, phi=0,
         noise_a=None, noise_f=None):
    """Generate harmonic signal.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    freq: Object
        Frequency of signal (Hz).
    phi: float
        Initial phase (radians).
    noise_a: callable
        Returns noise value added to amplitude.
    noise_f: callable
        Returns noise value added to frequency.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """

    times = np.arange(0, length, 1 / sample_rate)

    values = []
    for time_value in times:
        amp_value = amp
        if noise_a is not None:
            amp_value += noise_a(time_value)

        arg = _2PI * freq * time_value + phi
        if noise_f is not None:
            arg += noise_f(time_value)

        values.append(amp_value * cos(arg))

    return np.array(values), times


def amp_mod(length, sample_rate, func, freq, phi=0,
            noise_f=None, noise_a=None):
    """Amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    freq: float
        Frequency of signal (Hz).
    phi: float
        Initial phase (radians).
    func: Object
        Function that returns amplitude value depending on time.
    noise_f: Object
        Function that returns noise value added to frequency.
    noise_a: Object
        Function that returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """
    full_phase = phi
    delta_ph = _2PI * freq / sample_rate
    sampling_period = 1.0 / sample_rate
    times = np.arange(0, length + sampling_period, sampling_period)
    values = []
    for time_value in times:
        if noise_f is None:
            value = func(time_value) * cos(full_phase)
        else:
            value = func(time_value) * cos(full_phase + noise_f(time_value))
        if noise_a is not None:
            value += noise_a(time_value)
        values.append(value)
        full_phase += delta_ph
    values = np.array(values)
    return values, times


def freq_mod(length, sample_rate, amp, func, phi=0,
             noise_f=None, noise_a=None):
    """Amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    phi: float
        Initial phase (radians).
    func: Object
        Function that returns frequency values (in Hz) depending on time.
    noise_f: Object
        Function that returns noise value added to frequency.
    noise_a: Object
        Function that returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Full phase values.
    : np.array
        Time values.
    """
    full_phase = phi
    times = np.arange(0, length + 1.0 / sample_rate, 1.0 / sample_rate)
    values = []
    phs = []
    for time_value in times:
        arg = full_phase
        if noise_f is not None:
            arg += noise_f(time_value)
        value = amp * cos(arg)
        if noise_a is not None:
            value += noise_a(time_value)
        values.append(value)
        phs.append(full_phase)
        delta_ph = _2PI * func(time_value) / sample_rate
        full_phase += delta_ph
    values = np.array(values)
    phs = np.array(phs)
    return values, phs, times


def phase_mod(length, sample_rate, amp, freq, func,
              noise_f=None, noise_a=None):
    """Phase modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    freq: float
        Frequency of signal (Hz).
    func: Object
        Function that returns phase values (in radians) depending on time.
    noise_f: Object
        Function that returns noise value added to frequency.
    noise_a: Object
        Function that returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """
    sampling_period = 1.0 / sample_rate
    times = np.arange(0, length + sampling_period, sampling_period)
    values = []
    for time_value in times:
        arg = _2PI * freq * time_value + func(time_value)
        if noise_f is not None:
            arg += noise_f(time_value)
        value = amp * cos(arg)
        if noise_a is not None:
            value += noise_a(time_value)
        values.append(value)
    values = np.array(values)
    return values, times


def freq_amp_mod(length, sample_rate, a_func, f_func, phi=0):
    """Simultaneous frequency and amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    a_func: Object
        Function that returns amplitude value depending on time.
    f_func: Object
        Function that returns frequency values (in Hz) depending on time.
    phi: float
        Initial phase (radians).

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Full phase values.
    : np.array
        Time values.
    """
    full_phase = phi
    sampling_period = 1.0 / sample_rate
    times = np.arange(0, length + sampling_period, sampling_period)
    values = []
    phs = []
    for time_value in times:
        arg = full_phase
        values.append(a_func(time_value) * cos(arg))
        delta_ph = _2PI * f_func(time_value) / sample_rate
        phs.append(full_phase)
        full_phase += delta_ph
    values = np.array(values)
    phs = np.array(phs)
    return values, phs, times


def iq_demod(xdata, tdata, f_central, a_coeffs, b_coeffs):
    """Return instantaneous frequency of modulated signal using IQ processign.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    tdata: array_like
        Time values.
    f_central: float
        Carrier frequency.
    a_coeffs: array_like
        a values of filter.
    b_coeffs: array_like
        b values of filter.

    Returns
    -------
    : np.ndarray of floats
        Instantaneous frequency values.
    : np.ndarray
        Time values.
    """
    muli = xdata * np.cos(_2PI * f_central * tdata)
    mulq = xdata * np.sin(_2PI * f_central * tdata)
    muli_low = sig.lfilter(b_coeffs, a_coeffs, muli)
    mulq_low = sig.lfilter(b_coeffs, a_coeffs, mulq)
    analytic = muli_low + 1j * mulq_low
    phase = -unwrap(angle(analytic))
    freq = diff(phase) / _2PI / (tdata[1] - tdata[0]) + f_central
    return freq, tdata[:-1]


def envelope_by_extremums(xdata, sample_rate=1, tdata=None):
    """Calculate envelope by local extremums of signals.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    sample_rate: float
        Sampling frequency.
    tdata: array_like
        Time values. Use it for unregular discretized input signal.

    Returns
    --------
    : np.array
        Damping values.
    : np.array
        Time values.
    """
    if tdata is None:
        tdata = np.linspace(0, (len(xdata)-1)/sample_rate, len(xdata))
    t_new = []
    x_new = []
    xabs = abs(xdata)
    for x_l, x_c, x_r, t_c in zip(xabs[:-2], xabs[1:-1],
                                  xabs[2:], tdata[1:-1]):
        if (x_l < x_c) and (x_c >= x_r):
            t_new.append(t_c)
            x_new.append(x_c)
    if xabs[-1] > xabs[-2]:
        t_new.append(tdata[-1])
        x_new.append(xabs[-1])
    return np.array(x_new), np.array(t_new)


def digital_hilbert_filter(ntaps=101, window='hamming'):
    """Calculate digital hilbert filter.

    Parameters
    ----------
    ntaps: integer
        Length of filter.
    window: str
        Window. Default is 'hamming'.

    Returns
    -------
    : np.array
        Filter.
    """
    if ntaps % 2 == 0:
        raise ValueError("ntaps of digital Hilbert filter must be odd.")

    coeffs = np.zeros(ntaps)
    num = ntaps // 2
    for k in range(1, num+1, 2):
        coeffs[num + k] = 2 / PI / k
        coeffs[num - k] = -2 / PI / k

    wind = sig.get_window(window, ntaps)
    coeffs *= wind
    return coeffs


def freq_by_extremums(xdata, sample_rate):
    """Calculate frequency of oscillating signal by extremums.

    Parameters
    ----------
    xdata: array_like
        Values of input signals.
    sample_rate: float
        Sampling frequency (Hz).

    Returns
    -------
    : float
        Frequency.
    """

    if len(xdata) < 3:
        raise ValueError("Short signal")

    T = len(xdata) / sample_rate

    max_total = 0
    min_total = 0
    for prev, curr, nxt in zip(xdata[:-2], xdata[1:-1], xdata[2:]):
        if (prev < curr) and (curr >= nxt):
            max_total += 1
        if (prev > curr) and (curr <= nxt):
            min_total += 1

    if xdata[0] > xdata[1]:
        max_total += 1

    if xdata[0] < xdata[1]:
        min_total += 1

    return (max_total + min_total) / 2 / T


def freq_by_zeros(xdata, sample_rate):
    """Calculate average frequency of detrended oscillating signal by
    counting zeros."""

    T = len(xdata) / sample_rate

    zeros_total = 0
    for prev, curr in zip(xdata[:-1], xdata[1:]):
        if prev * curr < 0:
            zeros_total += 1
        elif prev != 0 and curr == 0:
            zeros_total += 1

    return zeros_total / 2 / T


def linint(xdata, tdata, ts_new):
    """Find values of xdata in ts_new points.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.
    ts_new: np.ndarray
        New time values.

    Returns
    -------
    : np.ndarray
        New signal values.
    """
    x_new = np.zeros(len(ts_new)) * np.nan
    for x_p, t_p, x_c, t_c in zip(xdata[:-1], tdata[:-1],
                                  xdata[1:], tdata[1:]):
        k = (x_c - x_p) / (t_c - t_p)
        b = x_p - k*t_p
        ind = (ts_new >= t_p) & (ts_new <= t_c)
        x_new[ind] = k * ts_new[ind] + b
    return x_new


def wave_lens(xdata, tdata):
    """Calculate wave lengths of signal by space between zeros.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.

    Returns
    -------
    : np.ndarray
        Wave lengths.
    : np.ndarray
        Time values.
    """
    tms = []
    for t_c, x_p, x_c in zip(tdata[1:], xdata[:-1], xdata[1:]):
        if x_p * x_c < 0:
            tms.append(t_c)
    lens = np.diff(tms) * 2
    t_lens = np.array(tms[1:])
    return lens, t_lens


def freqs_by_wave_len(xdata, tdata, cut_nans=True):
    """Calculate frequencies using lenghs of waves and linear
    interpolation.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.
    cut_nans: boolean
        If True, the nan values at the ends of the of the produced
        array will removed.

    Returns
    -------
    : np.ndarray
        Freqs values.
    """
    wls, t_wl = wave_lens(xdata, tdata)
    freqs = 1 / linint(wls, t_wl, tdata)
    if cut_nans:
        freqs_cut = []
        t_cut = []
        for (f, tt) in zip(freqs, tdata):
            if f == f:
                freqs_cut.append(f)
                t_cut.append(tt)
        return np.array(freqs_cut), t_cut
    return freqs



# Copyright (C) 2017-2020 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the base classes for Activities."""

import logging
from warnings import warn
from dsplab.helpers import import_entity

LOG = logging.getLogger(__name__)


class ActivityMeta(type):
    """Metaclass for Activity."""
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._class_info = {}
        try:
            cls._class_info['doc'] = attrs['__doc__']
        except KeyError:
            cls._class_info['doc'] = ""
        cls._class_info['class'] = name

    def class_info(cls):
        """Return the information about activity.

        Returns
        -------
        : dict
            Information about class of activity.
        """
        return cls._class_info

    def __call__(cls, *args, **kwargs):
        res = type.__call__(cls, *args, **kwargs)
        setattr(res, "class_info", cls.class_info)
        return res


class Activity(metaclass=ActivityMeta):
    """Any activity is the something that may be called and can
    provide the information about itself. To get working activity the
    __call__ method must be implemented."""
    def __call__(self, *args, **kwargs):
        """Call activity."""
        raise NotImplementedError

    def set_descr(self, descr):
        """Deprecated."""
        warn("Activity.set_descr() is deprecated. Use it for works only.")

    def info(self, as_string=None):
        """Deprecated."""
        warn("info() is deprecated and returns stub, don't use it")
        return self.__class__.__name__


class Worker(Activity):
    """Deprecated."""
    def __init__(self):
        super().__init__()
        warn("Worker is deprecated. Use Activity instead.")

    def add_param(self, name, value=None):
        """Deprecated."""
        warn("Worker.add_param() is deprecated. Don't use it.")
        setattr(self, name, value)

    def _reg_param(self, name):
        """Deprecated."""
        warn("Worker._reg_param() is deprecated. Don't use it.")


class Work(Activity):
    """Work is data processing that can be done in a variety of ways."""
    def __init__(self, descr=None, worker=None):
        super().__init__()
        self.set_descr(descr)
        self.set_worker(worker)

    def set_descr(self, text):
        """Set description."""
        self._descr = text

    def get_descr(self):
        """Return description."""
        return self._descr

    descr = property(get_descr, set_descr, doc="Description of work")

    def set_worker(self, act):
        """Set worker for doing work. Worker must be callable."""
        self._worker = act

    def __call__(self, *args, **kwargs):
        """Do work."""
        res = self._worker(*args, **kwargs)
        return res


def get_work_from_dict(settings, params=None):
    """Create and return Work instance described in dictionary."""
    if 'descr' in settings:
        descr = settings['descr']
    else:
        descr = ""

    if 'worker' not in settings:
        raise RuntimeError("No worker in settings")
    worker_settings = settings['worker']

    if 'class' in worker_settings.keys():
        key = 'class'
    elif 'function' in worker_settings.keys():
        key = 'function'
    else:
        raise RuntimeError("Work must be 'class' or 'function'")

    worker_name = worker_settings[key]

    if 'params' in worker_settings.keys():
        worker_params = worker_settings['params'].copy()

        for key in worker_params:
            if isinstance(worker_params[key], str):
                if worker_params[key]:
                    if worker_params[key][0] == "$":
                        params_key = worker_params[key][1:]
                        try:
                            worker_params[key] = params[params_key]
                        except KeyError:
                            msg = "${} not found in params".format(params_key)
                            raise RuntimeError(msg)

        worker = import_entity(worker_name)(**worker_params)
    else:
        if key == 'class':
            worker = import_entity(worker_name)()
        else:
            worker = import_entity(worker_name)

    work = Work(descr, worker)
    return work
